@@ -1,8 +1,11 @@
 import inspect
 
 import argparse
+import io
 import logging
+import os
 import tempfile
+from contextlib import redirect_stdout
 
 import cbor2
 import enum
@@ -48,34 +51,43 @@ def parse_uplc_param(param: str):
             return uplc.ast.data_from_json_dict(json.loads(param))
         except json.JSONDecodeError as e:
             raise ValueError(
-                f"Invalid parameter for contract passed, expected json value, got {param}"
+                f"Invalid parameter for contract passed, expected JSON value, got {param}"
+            ) from e
+        except Exception as e:
+            raise ValueError(
+                f"Expected parameter for contract to be in valid Plutus data JSON format, got {param}"
             ) from e
     else:
         try:
             return uplc.ast.data_from_cbor(bytes.fromhex(param))
-        except ValueError as e:
+        except Exception as e:
             raise ValueError(
                 "Expected hexadecimal CBOR representation of plutus datum but could not transform hex string to bytes."
             ) from e
 
 
 def parse_plutus_param(annotation, param: str):
-    if param.startswith("{"):
-        try:
-            param_dict = json.loads(param)
-        except json.JSONDecodeError as e:
-            raise ValueError(
-                f"Invalid parameter for contract passed, expected json value, got {param}"
-            ) from e
-        return plutus_data_from_json(annotation, param_dict)
-    else:
-        try:
-            param_bytes = bytes.fromhex(param)
-        except ValueError as e:
-            raise ValueError(
-                "Expected hexadecimal CBOR representation of plutus datum but could not transform hex string to bytes."
-            ) from e
-        return plutus_data_from_cbor(annotation, param_bytes)
+    try:
+        if param.startswith("{"):
+            try:
+                param_dict = json.loads(param)
+            except json.JSONDecodeError as e:
+                raise ValueError(
+                    f"Invalid parameter for contract passed, expected json value, got {param}"
+                ) from e
+            return plutus_data_from_json(annotation, param_dict)
+        else:
+            try:
+                param_bytes = bytes.fromhex(param)
+            except ValueError as e:
+                raise ValueError(
+                    "Expected hexadecimal CBOR representation of plutus datum but could not transform hex string to bytes."
+                ) from e
+            return plutus_data_from_cbor(annotation, param_bytes)
+    except ValueError as e:
+        raise ValueError(
+            f"Could not parse parameter {param} as type {annotation}. Please provide the parameter either as JSON or CBOR (in hexadecimal notation). Detailed error: {e}"
+        ) from e
 
 
 def plutus_data_from_json(annotation: typing.Type, x: dict):
@@ -103,7 +115,7 @@ def plutus_data_from_json(annotation: typing.Type, x: dict):
                 for ann in annotation.__dict__["__args__"]:
                     try:
                         return plutus_data_from_json(ann, x)
-                    except pycardano.DeserializeException:
+                    except (pycardano.DeserializeException, KeyError, ValueError):
                         pass
                 raise ValueError(
                     f"Could not find matching type for {x} in {annotation}"
@@ -139,8 +151,15 @@ def plutus_data_from_json(annotation: typing.Type, x: dict):
 def plutus_data_from_cbor(annotation: typing.Type, x: bytes):
     try:
         if annotation in (int, bytes):
-            return cbor2.loads(x)
+            res = cbor2.loads(x)
+            if not isinstance(res, annotation):
+                raise ValueError(
+                    f"Expected {annotation} but got {type(x)} from {x.hex()}"
+                )
+            return res
         if annotation is None:
+            if not x == cbor2.dumps(None):
+                raise ValueError(f"Expected None but got {x.hex()}")
             return None
         if isinstance(annotation, typing._GenericAlias):
             # Annotation is a List or Dict
@@ -155,14 +174,14 @@ def plutus_data_from_cbor(annotation: typing.Type, x: bytes):
                 return {
                     plutus_data_from_cbor(
                         annotation_key, cbor2.dumps(k)
-                    ): plutus_data_from_cbor(annotation_val, v)
+                    ): plutus_data_from_cbor(annotation_val, cbor2.dumps(v))
                     for k, v in cbor2.loads(x).items()
                 }
             if annotation.__origin__ == typing.Union:
                 for ann in annotation.__dict__["__args__"]:
                     try:
                         return plutus_data_from_cbor(ann, x)
-                    except pycardano.DeserializeException:
+                    except (pycardano.DeserializeException, ValueError):
                         pass
                 raise ValueError(
                     f"Could not find matching type for {x.hex()} in {annotation}"
@@ -255,7 +274,13 @@ def perform_command(args):
         with tmp_input_file.open("w") as fp:
             fp.write(source_code)
         sys.path.append(str(pathlib.Path(tmp_input_file).parent.absolute()))
-        sc = importlib.import_module(pathlib.Path(tmp_input_file).stem)
+        try:
+            sc = importlib.import_module(pathlib.Path(tmp_input_file).stem)
+        except Exception as e:
+            # replace the traceback with an error pointing to the input file
+            raise SyntaxError(
+                f"Could not import the input file as python module. Make sure the input file is valid python code. Error: {e}",
+            ) from e
         sys.path.pop()
     # load the passed parameters if not a lib
     if purpose == Purpose.lib:
@@ -266,8 +291,17 @@ def perform_command(args):
         try:
             argspec = inspect.signature(sc.validator)
         except AttributeError:
+            can_command_with_lib = command in (
+                Command.compile,
+                Command.build,
+                Command.compile_pluto,
+                Command.parse,
+                Command.lint,
+            )
+            command_with_lib = f" or {command.value} using `opshin {command.value} lib {str(input_file)}`."
             raise AssertionError(
-                f"Contract has no function called 'validator'. Make sure the compiled contract contains one function called 'validator' or {command.value} using `opshin {command.value} lib {str(input_file)}`."
+                f"Contract has no function called 'validator'. Make sure the compiled contract contains one function called 'validator'"
+                + (command_with_lib if can_command_with_lib else ".")
             )
         annotations = [
             (x.name, x.annotation or prelude.Anything)
@@ -281,9 +315,19 @@ def perform_command(args):
         parsed_params = []
         uplc_params = []
         for i, (c, a) in enumerate(zip(annotations, args.args)):
-            uplc_param = parse_uplc_param(a)
+            try:
+                uplc_param = parse_uplc_param(a)
+            except ValueError as e:
+                raise ValueError(
+                    f"Could not parse parameter {i} ('{a}') as UPLC data. Please provide the parameter either as JSON or CBOR (in hexadecimal notation). Detailed error: {e}"
+                ) from None
             uplc_params.append(uplc_param)
-            param = parse_plutus_param(c[1], a)
+            try:
+                param = parse_plutus_param(c[1], a)
+            except ValueError as e:
+                raise ValueError(
+                    f"Could not parse parameter {i} ('{a}') as type {c[1]}. Please provide the parameter either as JSON or CBOR (in hexadecimal notation). Detailed error: {e}"
+                ) from None
             parsed_params.append(param)
         onchain_params, param_types = check_params(
             command,
@@ -293,18 +337,20 @@ def perform_command(args):
             parsed_params,
             compiler_config.force_three_params,
         )
+        assert (
+            onchain_params or purpose == Purpose.lib
+        ), "The validator function must have at least one on-chain parameter for non-library contracts."
 
+    py_ret = Command.eval
     if command == Command.eval:
         assert purpose != Purpose.lib, "Can not evaluate a library"
-        print("Starting execution")
-        print("------------------")
-        try:
-            ret = sc.validator(*parsed_params)
-        except Exception as e:
-            print(f"Exception of type {type(e).__name__} caused")
-            ret = e
-        print("------------------")
-        print(ret)
+        print("Python execution started")
+        with redirect_stdout(open(os.devnull, "w")):
+            try:
+                py_ret = sc.validator(*parsed_params)
+            except Exception as e:
+                py_ret = e
+        command = Command.eval_uplc
 
     source_ast = compiler.parse(source_code, filename=input_file)
 
@@ -382,6 +428,7 @@ Note that opshin errors may be overly restrictive as they aim to prevent code wi
         if purpose == Purpose.lib:
             script_arts = PlutusContract(
                 built_code,
+                title=pathlib.Path(input_file).stem,
             )
         else:
             script_arts = PlutusContract(
@@ -390,25 +437,34 @@ Note that opshin errors may be overly restrictive as they aim to prevent code wi
                 redeemer_type=onchain_params[1 if len(onchain_params) == 3 else 0],
                 parameter_types=param_types,
                 purpose=(purpose,),
+                title=pathlib.Path(input_file).stem,
             )
         script_arts.dump(target_dir)
 
         print(f"Wrote script artifacts to {target_dir}/")
         return
     if command == Command.eval_uplc:
-        print("Starting execution")
-        print("------------------")
+        print("UPLC execution started")
         assert isinstance(code, uplc.ast.Program)
         raw_ret = uplc.eval(code)
+        print("------LOGS--------")
+        if raw_ret.logs:
+            for log in raw_ret.logs:
+                print(" > " + log)
+        else:
+            print("No logs")
+        print("------COST--------")
+        print(f"CPU: {raw_ret.cost.cpu} | MEM: {raw_ret.cost.memory}")
         if isinstance(raw_ret.result, Exception):
-            print("An exception was raised")
+            print("----EXCEPTION-----")
             ret = raw_ret.result
         else:
-            print("Execution succeeded")
+            print("-----SUCCESS------")
             ret = uplc.dumps(raw_ret.result)
-            print(f"CPU: {raw_ret.cost.cpu} | MEM: {raw_ret.cost.memory}")
-        print("------------------")
-        print(ret)
+        print(str(ret), end="")
+        if not isinstance(py_ret, Command):
+            print(" (Python: " + str(py_ret) + ")", end="")
+        print()
 
 
 def parse_args():
@@ -503,8 +559,8 @@ def parse_args():
     a.add_argument(
         f"-O",
         type=int,
-        help=f"Optimization level from 0 (no optimization) to 3 (aggressive optimization, removes traces). Defaults to 1.",
-        default=1,
+        help=f"Optimization level from 0 (no optimization) to 3 (aggressive optimization). Defaults to 1.",
+        default=2,
         choices=range(len(OPT_CONFIGS)),
         dest="opt_level",
     )
