@@ -1,10 +1,42 @@
+import ast
 import copy
+import typing
+from ast import Load, Name, Constant, Slice
 
+import pluthon as plt
+import uplc.ast as uplc
+from pycardano import PlutusData
 from uplc.ast import data_from_cbor
 
+from .bridge import to_uplc_builtin
+from .prelude import Nothing
+from .type_impls import (
+    InstanceType,
+    UnionType,
+    UnitType,
+    RecordType,
+    transform_ext_params_map,
+    AnyType,
+    transform_output_map,
+    ClassType,
+    PolymorphicFunctionInstanceType,
+    ListType,
+    TupleType,
+    PairType,
+    IntegerInstanceType,
+    empty_list,
+    DictType,
+    ByteStringType,
+    FunctionType,
+    OUnit,
+)
+from .type_inference import map_to_orig_name, AggressiveTypeInferencer
+from .typed_ast import *
+
 from .compiler_config import DEFAULT_CONFIG
 from .optimize.optimize_const_folding import OptimizeConstantFolding
-from .optimize.optimize_remove_comments import OptimizeRemoveDeadconstants
+from .optimize.optimize_remove_deadconstants import OptimizeRemoveDeadconstants
+from .optimize.optimize_union_expansion import OptimizeUnionExpansion
 
 from .rewrite.rewrite_assert_none import RewriteAssertNone
 from .rewrite.rewrite_augassign import RewriteAugAssign
@@ -29,21 +61,27 @@ from .rewrite.rewrite_subscript38 import RewriteSubscript38
 from .rewrite.rewrite_tuple_assign import RewriteTupleAssign
 from .optimize.optimize_remove_pass import OptimizeRemovePass
 from .optimize.optimize_remove_deadvars import OptimizeRemoveDeadvars, NameLoadCollector
-from .type_inference import *
 from .util import (
     CompilingNodeTransformer,
     NoOp,
-)
-from .typed_ast import (
-    transform_ext_params_map,
-    transform_output_map,
-    RawPlutoExpr,
+    OVar,
+    OLambda,
+    OLet,
+    OPSHIN_LOGGER,
+    all_vars,
+    SafeOLambda,
+    opshin_name_scheme_compatible_varname,
+    force_params,
+    SafeApply,
+    SafeLambda,
+    written_vars,
+    custom_fix_missing_locations,
 )
 
 
 BoolOpMap = {
-    And: plt.And,
-    Or: plt.Or,
+    ast.And: plt.And,
+    ast.Or: plt.Or,
 }
 
 
@@ -69,6 +107,9 @@ def rec_constant_map_data(c):
                 )
             )
         )
+    # This can occur when PlutusData is generated during constant folding
+    if isinstance(c, PlutusData):
+        return data_from_cbor(c.to_cbor())
     raise NotImplementedError(f"Unsupported constant type {type(c)}")
 
 
@@ -95,6 +136,7 @@ def rec_constant_map(c):
                 )
             ]
         )
+    # This can occur when PlutusData is generated during constant folding
     if isinstance(c, PlutusData):
         return data_from_cbor(c.to_cbor())
     raise NotImplementedError(f"Unsupported constant type {type(c)}")
@@ -123,7 +165,7 @@ def wrap_validator_double_function(x: plt.AST, pass_through: int = 0):
                 # call the validator with a0, a1, and plug in "Nothing" for data
                 plt.Apply(
                     OVar("p"),
-                    plt.UPLCConstant(uplc.PlutusConstr(6, [])),
+                    plt.UPLCConstant(to_uplc_builtin(Nothing())),
                     OVar("a0"),
                     OVar("a1"),
                 ),
@@ -215,7 +257,7 @@ class PlutoCompiler(CompilingNodeTransformer):
             main_fun: typing.Optional[InstanceType] = None
             for s in node.body:
                 if (
-                    isinstance(s, FunctionDef)
+                    isinstance(s, ast.FunctionDef)
                     and s.orig_name == self.validator_function_name
                 ):
                     main_fun = s
@@ -257,10 +299,10 @@ class PlutoCompiler(CompilingNodeTransformer):
                 [
                     TypedReturn(
                         TypedCall(
-                            func=Name(
+                            func=ast.Name(
                                 id=main_fun.name,
                                 typ=InstanceType(main_fun_typ),
-                                ctx=Load(),
+                                ctx=ast.Load(),
                             ),
                             typ=main_fun_typ.rettyp,
                             args=[
@@ -295,9 +337,7 @@ class PlutoCompiler(CompilingNodeTransformer):
                         )
                         for x in all_vs
                     ],
-                    self.visit_sequence(body)(
-                        plt.ConstrData(plt.Integer(0), plt.EmptyDataList())
-                    ),
+                    self.visit_sequence(body)(OUnit),
                 ),
             )
             self.current_function_typ.pop()
@@ -326,15 +366,13 @@ class PlutoCompiler(CompilingNodeTransformer):
                     )
                     for x in all_vs
                 ],
-                self.visit_sequence(body)(
-                    plt.ConstrData(plt.Integer(0), plt.EmptyDataList())
-                ),
+                self.visit_sequence(body)(OUnit),
             )
 
         cp = plt.Program((1, 0, 0), validator)
         return cp
 
-    def visit_Constant(self, node: TypedConstant) -> plt.AST:
+    def visit_Constant(self, node: Constant) -> plt.AST:
         if isinstance(node.value, bytes) and node.value != b"":
             try:
                 bytes.fromhex(node.value.decode())
@@ -378,7 +416,7 @@ class PlutoCompiler(CompilingNodeTransformer):
             x,
         )
 
-    def visit_AnnAssign(self, node: AnnAssign) -> CallAST:
+    def visit_AnnAssign(self, node: TypedAnnAssign) -> CallAST:
         assert isinstance(
             node.target, Name
         ), "Assignments to other things then names are not supported"
@@ -408,7 +446,7 @@ class PlutoCompiler(CompilingNodeTransformer):
             x,
         )
 
-    def visit_Name(self, node: TypedName) -> plt.AST:
+    def visit_Name(self, node: Name) -> plt.AST:
         # depending on load or store context, return the value of the variable or its name
         if not isinstance(node.ctx, Load):
             raise NotImplementedError(f"Context {node.ctx} not supported")
@@ -479,7 +517,7 @@ class PlutoCompiler(CompilingNodeTransformer):
         body = node.body.copy()
         # defaults to returning None if there is no return statement
         if node.typ.typ.rettyp.typ == AnyType():
-            ret_val = plt.ConstrData(plt.Integer(0), plt.EmptyDataList())
+            ret_val = OUnit
         else:
             ret_val = plt.Unit()
         read_vs = sorted(list(node.typ.typ.bound_vars.keys()))
@@ -508,7 +546,7 @@ class PlutoCompiler(CompilingNodeTransformer):
         # by overwriting them with arguments to its self-recall
         if node.orelse:
             # If there is orelse, transform it to an appended sequence (TODO check if this is correct)
-            cn = copy(node)
+            cn = copy.copy(node)
             cn.orelse = []
             return self.visit_sequence([cn] + node.orelse)
         compiled_c = self.visit(node.test)
@@ -523,7 +561,7 @@ class PlutoCompiler(CompilingNodeTransformer):
                     plt.Apply(
                         OVar("while"),
                         OVar("while"),
-                        *deepcopy(pwritten_vs),
+                        *copy.deepcopy(pwritten_vs),
                     )
                 ),
                 x,
@@ -535,16 +573,18 @@ class PlutoCompiler(CompilingNodeTransformer):
                 ("adjusted_next", SafeLambda(written_vs, x)),
                 (
                     "while",
-                    s_fun(SafeApply(OVar("adjusted_next"), *deepcopy(pwritten_vs))),
+                    s_fun(
+                        SafeApply(OVar("adjusted_next"), *copy.deepcopy(pwritten_vs))
+                    ),
                 ),
             ],
-            plt.Apply(OVar("while"), OVar("while"), *deepcopy(pwritten_vs)),
+            plt.Apply(OVar("while"), OVar("while"), *copy.deepcopy(pwritten_vs)),
         )
 
     def visit_For(self, node: TypedFor) -> CallAST:
         if node.orelse:
             # If there is orelse, transform it to an appended sequence (TODO check if this is correct)
-            cn = copy(node)
+            cn = copy.copy(node)
             cn.orelse = []
             return self.visit_sequence([cn] + node.orelse)
         assert isinstance(node.iter.typ, InstanceType)
@@ -572,7 +612,7 @@ class PlutoCompiler(CompilingNodeTransformer):
                                 OVar("for"),
                                 OVar("for"),
                                 plt.TailList(OVar("iter")),
-                                *deepcopy(pwritten_vs),
+                                *copy.deepcopy(pwritten_vs),
                             )
                         ),
                     ),
@@ -587,7 +627,7 @@ class PlutoCompiler(CompilingNodeTransformer):
                             plt.Apply(
                                 OVar("adjusted_next"),
                                 plt.Var(node.target.id),
-                                *deepcopy(pwritten_vs),
+                                *copy.deepcopy(pwritten_vs),
                             )
                         ),
                     ),
@@ -596,7 +636,7 @@ class PlutoCompiler(CompilingNodeTransformer):
                     OVar("for"),
                     OVar("for"),
                     compiled_iter,
-                    *deepcopy(pwritten_vs),
+                    *copy.deepcopy(pwritten_vs),
                 ),
             )
         raise NotImplementedError(
@@ -611,10 +651,10 @@ class PlutoCompiler(CompilingNodeTransformer):
             plt.Ite(
                 self.visit(node.test),
                 self.visit_sequence(node.body)(
-                    SafeApply(OVar("adjusted_next"), *deepcopy(pwritten_vs))
+                    SafeApply(OVar("adjusted_next"), *copy.deepcopy(pwritten_vs))
                 ),
                 self.visit_sequence(node.orelse)(
-                    SafeApply(OVar("adjusted_next"), *deepcopy(pwritten_vs))
+                    SafeApply(OVar("adjusted_next"), *copy.deepcopy(pwritten_vs))
                 ),
             ),
         )
@@ -628,9 +668,6 @@ class PlutoCompiler(CompilingNodeTransformer):
             value_plt = transform_output_map(node.value.typ)(value_plt)
         return lambda _: value_plt
 
-    def visit_Pass(self, node: TypedPass) -> CallAST:
-        return self.visit_sequence([])
-
     def visit_Subscript(self, node: TypedSubscript) -> plt.AST:
         assert isinstance(
             node.value.typ, InstanceType
@@ -664,7 +701,7 @@ class PlutoCompiler(CompilingNodeTransformer):
             assert isinstance(node.ctx, Load), "Pairs are read-only"
             assert (
                 0 <= index < 2
-            ), f"Pairs only have 2 elements, index should be 0 or 1, is {node.slice.value}"
+            ), f"Pairs only have 2 elements, index should be -2, -1, 0 or 1, found {node.slice.value}"
             member_func = plt.FstPair if index == 0 else plt.SndPair
             # the content of pairs is always Data, so we need to unwrap
             member_typ = node.typ
@@ -714,6 +751,12 @@ class PlutoCompiler(CompilingNodeTransformer):
                     ),
                 )
             else:
+                assert (
+                    node.slice.upper is not None
+                ), "Only slices with upper bound supported"
+                assert (
+                    node.slice.lower is not None
+                ), "Only slices with lower bound supported"
                 return OLet(
                     [
                         (
@@ -781,7 +824,9 @@ class PlutoCompiler(CompilingNodeTransformer):
                     [
                         (
                             "key",
-                            self.visit(node.slice),
+                            transform_output_map(node.slice.typ)(
+                                self.visit(node.slice),
+                            ),
                         )
                     ],
                     transform_ext_params_map(dict_typ.value_typ)(
@@ -791,9 +836,7 @@ class PlutoCompiler(CompilingNodeTransformer):
                                 OLambda(
                                     ["x"],
                                     plt.EqualsData(
-                                        transform_output_map(node.slice.typ)(
-                                            OVar("key")
-                                        ),
+                                        OVar("key"),
                                         plt.FstPair(OVar("x")),
                                     ),
                                 ),
@@ -804,6 +847,19 @@ class PlutoCompiler(CompilingNodeTransformer):
                 )
         elif isinstance(node.value.typ.typ, ByteStringType):
             if not isinstance(node.slice, Slice):
+                if isinstance(node.slice, Constant) and node.slice.value >= 0:
+                    return plt.IndexByteString(
+                        self.visit(node.value),
+                        self.visit(node.slice),
+                    )
+                elif isinstance(node.slice, Constant) and node.slice.value < 0:
+                    return plt.IndexByteString(
+                        self.visit(node.value),
+                        plt.AddInteger(
+                            self.visit(node.slice),
+                            plt.LengthOfByteString(self.visit(node.value)),
+                        ),
+                    )
                 return OLet(
                     [
                         (
@@ -1078,12 +1134,25 @@ class PlutoCompiler(CompilingNodeTransformer):
             joined_str = plt.AppendString(self.visit(v), joined_str)
         return joined_str
 
-    def generic_visit(self, node: AST) -> plt.AST:
+    def generic_visit(self, node: TypedAST) -> plt.AST:
         raise NotImplementedError(f"Can not compile {node}")
 
 
+def parse(
+    source: str,
+    filename=None,
+) -> ast.AST:
+    """
+    Parse source code into an AST
+
+    Currently passes everything through Python's ast module.
+    """
+    tree = ast.parse(source, filename=filename)
+    return tree
+
+
 def compile(
-    prog: AST,
+    prog: ast.AST,
     filename=None,
     validator_function_name="validator",
     config=DEFAULT_CONFIG,
@@ -1094,6 +1163,7 @@ def compile(
         # Rewrites that simplify the python code
         RewriteForbiddenReturn(),
         OptimizeConstantFolding() if config.constant_folding else NoOp(),
+        OptimizeUnionExpansion() if config.expand_union_types else NoOp(),
         RewriteSubscript38(),
         RewriteAugAssign(),
         RewriteComparisonChaining(),
