@@ -13,13 +13,58 @@ security into the Smart Contract by checking type correctness.
 """
 
 import re
+import ast
+from ast import *
+import typing
+from collections import defaultdict
+from copy import copy
+from hashlib import sha256
 
+from frozenlist2 import frozenlist
+from ordered_set import OrderedSet
 from pycardano import PlutusData
 from typing import Union
+import pluthon as plt
 from .typed_ast import *
-from .util import CompilingNodeTransformer
+from .util import (
+    CompilingNodeTransformer,
+    distinct,
+    TypedNodeVisitor,
+    OPSHIN_LOGGER,
+    custom_fix_missing_locations,
+    read_vars,
+    externally_bound_vars,
+)
 from .fun_impls import PythonBuiltInTypes
 from .rewrite.rewrite_cast_condition import SPECIAL_BOOL
+from .type_impls import (
+    Type,
+    ByteStringType,
+    IntegerType,
+    StringType,
+    AnyType,
+    BoolType,
+    InstanceType,
+    RecordType,
+    PolymorphicFunctionType,
+    Record,
+    BoolInstanceType,
+    IntegerInstanceType,
+    UnitInstanceType,
+    ByteStringInstanceType,
+    StringInstanceType,
+    ListType,
+    DictType,
+    UnionType,
+    PairType,
+    TypeInferenceError,
+    UnitType,
+    ATOMIC_TYPES,
+    ClassType,
+    TupleType,
+    PolymorphicFunctionInstanceType,
+    FunctionType,
+)
 
 # from frozendict import frozendict
 
@@ -73,6 +118,22 @@ DUNDER_MAP = {
     ast.Or: "__or__",
 }
 
+DUNDER_REVERSE_MAP = {
+    ast.Add: "__radd__",
+    ast.Sub: "__rsub__",
+    ast.Mult: "__rmul__",
+    ast.Div: "__rtruediv__",
+    ast.FloorDiv: "__rfloordiv__",
+    ast.Mod: "__rmod__",
+    ast.Pow: "__rpow__",
+    ast.LShift: "__rlshift__",
+    ast.RShift: "__rrshift__",
+    ast.And: "__rand__",
+    ast.Or: "__ror__",
+}
+
+ALL_DUNDERS = set(DUNDER_MAP.values()).union(set(DUNDER_REVERSE_MAP.values()))
+
 
 def record_from_plutusdata(c: PlutusData):
     return Record(
@@ -187,6 +248,8 @@ def intersection_types(*ts: Type):
 
 
 def find_max_type(elts: typing.List[InstanceType]):
+    if not elts:
+        return InstanceType(AnyType())
     set_elts = OrderedSet(elts)
     max_typ = None
     for m in elts:
@@ -258,7 +321,7 @@ class TypeCheckVisitor(TypedNodeVisitor):
         varname = node.args[0].id
         return ({varname: target_class}, {varname: union_without_target_class})
 
-    def visit_BoolOp(self, node: BoolOp) -> PairType:
+    def visit_BoolOp(self, node: BoolOp) -> TypeMapPair:
         res = {}
         inv_res = {}
         checks = [self.visit(v) for v in node.values]
@@ -278,7 +341,7 @@ class TypeCheckVisitor(TypedNodeVisitor):
                 if len(ts) < len(checks):
                     continue
                 inv_res[v] = union_types(*ts)
-        if isinstance(node.op, Or):
+        elif isinstance(node.op, Or):
             # a disjunction is just the union, but some type must be checked in every disjunction
             for v, ts in checked_types.items():
                 if len(ts) < len(checks):
@@ -287,9 +350,11 @@ class TypeCheckVisitor(TypedNodeVisitor):
             # if the disjunction fails, then it must be in the intersection of the inverses
             for v, ts in inv_checked_types.items():
                 inv_res[v] = intersection_types(*ts)
+        else:
+            raise NotImplementedError(f"Unsupported boolean operator {node.op}")
         return (res, inv_res)
 
-    def visit_UnaryOp(self, node: UnaryOp) -> PairType:
+    def visit_UnaryOp(self, node: UnaryOp) -> TypeMapPair:
         (res, inv_res) = self.visit(node.operand)
         if isinstance(node.op, Not):
             return (inv_res, res)
@@ -352,6 +417,13 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
                 "This can happen for example if you redefine a (renamed) imported function but try to use it before the redefinition."
             )
 
+    def is_defined_in_current_scope(self, name: str) -> bool:
+        try:
+            self.variable_type(name)
+            return True
+        except TypeInferenceError:
+            return False
+
     def enter_scope(self):
         self.scopes.append({})
 
@@ -381,44 +453,76 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
         operation = None
         args = []
         if isinstance(node, UnaryOp):
-            operand = node.operand
+            operand = self.visit(node.operand)
             operation = node.op
         elif isinstance(node, BinOp):
-            operand = node.left
+            operand = self.visit(node.left)
             operation = node.op
-            args.append(node.right)
+            args = [self.visit(node.right)]
         elif isinstance(node, Compare):
             operation = node.ops[0]
             if any([isinstance(operation, x) for x in [ast.In, ast.NotIn]]):
-                operand = node.comparators[0]
-                args = [node.left]
+                operand = self.visit(node.comparators[0])
+                args = [self.visit(node.left)]
             else:
-                operand = node.left
-                args = node.comparators
+                operand = self.visit(node.left)
+                args = [self.visit(c) for c in node.comparators]
             assert len(node.ops) == 1, "Only support one op at a time"
-        if operand is not None and hasattr(operand, "id"):
-            operand_type = self.variable_type(operand.id)
-            if (
-                operation.__class__ in DUNDER_MAP
-                and isinstance(operand_type, InstanceType)
-                and isinstance(operand_type.typ, RecordType)
-            ):
-                dunder = DUNDER_MAP[operation.__class__]
-                operand_class_name = operand_type.typ.record.name
-                method_name = f"{operand_class_name}_{dunder}"
-                if any([method_name in scope for scope in self.scopes]):
-                    call = ast.Call(
-                        func=ast.Attribute(
-                            value=operand,
-                            attr=dunder,
-                            ctx=ast.Load(),
-                        ),
-                        args=args,
-                        keywords=[],
+        operand_type = operand.typ
+        if (
+            operation.__class__ in DUNDER_MAP
+            and isinstance(operand_type, InstanceType)
+            and isinstance(operand_type.typ, RecordType)
+        ):
+            dunder = DUNDER_MAP[operation.__class__]
+            operand_class_name = operand_type.typ.record.name
+            method_name = f"{operand_class_name}_+_{dunder}"
+            if any([method_name in scope for scope in self.scopes]):
+                call = ast.Call(
+                    func=ast.Attribute(
+                        value=operand,
+                        attr=dunder,
+                        ctx=ast.Load(),
+                    ),
+                    args=args,
+                    keywords=[],
+                )
+                call.func.orig_id = f"{operand_class_name}.{dunder}"
+                call.func.id = method_name
+                call = self.visit_Call(call)
+                if (dunder == "__contains__" and isinstance(operation, ast.NotIn)) or (
+                    dunder == "__bool__" and isinstance(operation, ast.Not)
+                ):
+                    # we need to negate the result
+                    not_call = TypedUnaryOp(
+                        op=ast.Not(), operand=call, typ=BoolInstanceType
                     )
-                    call.func.orig_id = None
-                    call.func.id = method_name
-                    return self.visit_Call(call)
+                    return not_call
+                return call
+        # if this is not supported, try the reverse dunder
+        # note we assume 1, i.e. allow only a single right operand
+        right_op_typ = args[0].typ if len(args) == 1 else None
+        if (
+            operation.__class__ in DUNDER_REVERSE_MAP
+            and isinstance(right_op_typ, InstanceType)
+            and isinstance(right_op_typ.typ, RecordType)
+        ):
+            dunder = DUNDER_REVERSE_MAP[operation.__class__]
+            right_class_name = right_op_typ.typ.record.name
+            method_name = f"{right_class_name}_+_{dunder}"
+            if any([method_name in scope for scope in self.scopes]):
+                call = ast.Call(
+                    func=ast.Attribute(
+                        value=args[0],
+                        attr=dunder,
+                        ctx=ast.Load(),
+                    ),
+                    args=[operand],
+                    keywords=[],
+                )
+                call.func.orig_id = f"{right_class_name}.{dunder}"
+                call.func.id = method_name
+                return self.visit_Call(call)
         return None
 
     def type_from_annotation(self, ann: expr):
@@ -455,7 +559,16 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
                 ann.value, Name
             ), "Only Union, Dict and List are allowed as Generic types"
             if ann.value.orig_id == "Union":
-                for elt in ann.slice.elts:
+                if isinstance(ann.slice, Name):
+                    elts = [ann.slice]
+                elif isinstance(ann.slice, ast.Tuple):
+                    elts = ann.slice.elts
+                else:
+                    raise TypeInferenceError(
+                        "Union must combine several classes, use Union[Class1, Class2, ...]"
+                    )
+                # only allow List[Anything] and Dict[Anything, Anything] in unions
+                for elt in elts:
                     if isinstance(elt, Subscript) and elt.value.id == "List":
                         assert (
                             isinstance(elt.slice, Name)
@@ -465,10 +578,8 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
                         assert all(
                             isinstance(e, Name) and e.orig_id == "Anything"
                             for e in elt.slice.elts
-                        ), f"Only Dict[Anything, Anything] or Dict is supported in Unions. Received Dict[{elt.slice.elts[0].orig_id}, {elt.slice.elts[1].orig_id}]."
-                ann_types = frozenlist(
-                    [self.type_from_annotation(e) for e in ann.slice.elts]
-                )
+                        ), f"Only Dict[Anything, Anything] is supported in Unions. Received Dict[{elt.slice.elts[0].orig_id}, {elt.slice.elts[1].orig_id}]."
+                ann_types = frozenlist([self.type_from_annotation(e) for e in elts])
                 # flatten encountered union types
                 ann_types = frozenlist(
                     sum(
@@ -546,31 +657,52 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
                     continue
                 func = copy(attribute)
                 if func.name[0:2] == "__" and func.name[-2:] == "__":
-                    assert any(
-                        [func.name == value for key, value in DUNDER_MAP.items()]
-                    ), f"The following Dunder methods are supported {list(DUNDER_MAP.values())}. Received {func.name} which is not supported"
-                func.name = f"{n.name}_{attribute.name}"
-                for arg in func.args.args:
-                    if not arg.annotation is None:
-                        if isinstance(arg.annotation, ast.Name):
-                            assert (
-                                arg.annotation is None or arg.annotation.id != n.name
-                            ), "Invalid Python, class name is undefined at this stage."
-                        elif (
-                            isinstance(arg.annotation, ast.Subscript)
-                            and arg.annotation.value.id == "Union"
+                    assert (
+                        func.name in ALL_DUNDERS
+                    ), f"The following Dunder methods are supported {sorted(ALL_DUNDERS)}. Received {func.name} which is not supported"
+                func.name = f"{n.name}_+_{attribute.name}"
+
+                def does_literally_reference_self(arg):
+                    if arg is None:
+                        return False
+                    if isinstance(arg, Name) and arg.id == n.name:
+                        return True
+                    if (
+                        isinstance(arg, ast.Subscript)
+                        and isinstance(arg.value, Name)
+                        and arg.value.id in ("Union", "List", "Dict")
+                    ):
+                        # Only possible for List, Dict and Union
+                        if any(
+                            does_literally_reference_self(e) for e in arg.slice.elts
                         ):
-                            for s in arg.annotation.slice.elts:
-                                assert (
-                                    isinstance(s, Name) and s.id != n.name
-                                ) or isinstance(
-                                    s, Constant
-                                ), "Invalid Python, class name is undefined at this stage."
-                assert isinstance(func.returns, Constant) or (
-                    isinstance(func.returns, Name) and func.returns.id != n.name
-                ), "Invalid Python, class name is undefined at this stage"
+                            return True
+                    return False
+
+                for arg in func.args.args:
+                    assert not does_literally_reference_self(
+                        arg.annotation
+                    ), f"Argument '{arg.arg}' of method '{attribute.name}' in class '{n.name}' literally references the class itself. This is not allowed. If you want to reference the class itself, use 'Self' as type annotation."
+                assert not does_literally_reference_self(
+                    func.returns
+                ), f"Return type of method '{attribute.name}' in class '{n.name}' literally references the class itself. This is not allowed. If you want to reference the class itself, use 'Self' as type annotation."
                 ann = ast.Name(id=n.name, ctx=ast.Load())
+                if len(func.args.args) == 0:
+                    raise TypeError(
+                        f"Method '{attribute.orig_name}' in class '{n.orig_name}' must have at least one argument (self)"
+                    )
                 custom_fix_missing_locations(ann, attribute.args.args[0])
+                if func.args.args[0].orig_arg != "self":
+                    OPSHIN_LOGGER.warning(
+                        f"The first argument of method '{attribute.name}' in class '{n.orig_name}' should be named 'self', but found '{func.args.args[0].orig_arg}'. This is not enforced, but recommended."
+                    )
+                if func.args.args[0].annotation is not None and not (
+                    isinstance(func.args.args[0].annotation, Name)
+                    and func.args.args[0].annotation.id == "Self"
+                ):
+                    raise TypeError(
+                        f"The first argument of method '{attribute.name}' in class '{n.name}' must either not be annotated or be annotated with 'Self' to indicate that it is the instance of the class."
+                    )
                 ann.orig_id = attribute.args.args[0].orig_arg
                 func.args.args[0].annotation = ann
                 additional_functions.append(func)
@@ -596,7 +728,7 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
         return stmts
 
     def visit_ClassDef(self, node: ClassDef) -> TypedClassDef:
-        class_record = RecordReader.extract(node, self)
+        class_record = RecordReader(self).extract(node)
         typ = RecordType(class_record)
         self.set_variable_type(node.name, typ)
         self.FUNCTION_ARGUMENT_REGISTRY[node.name] = [
@@ -659,7 +791,7 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
         for t in node.targets:
             assert isinstance(
                 t, Name
-            ), "Can only assign to variable names, no type deconstruction"
+            ), "Can only assign to variable names (e.g., x = 5). OpShin does not allow assigning to tuple deconstructors (e.g., a, b = (1, 2)) or to dicts, lists, or members (e.g., x[0] = 1; x.foo = 1)"
             # Check compatability to previous types -> variable can be bound in a function before and needs to maintain type
             self.set_variable_type(t.id, typed_ass.value.typ)
         typed_ass.targets = [self.visit(t) for t in node.targets]
@@ -823,8 +955,6 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
     def visit_Compare(self, node: Compare) -> Union[TypedCompare, TypedCall]:
         dunder_node = self.dunder_override(node)
         if dunder_node is not None:
-            if isinstance(node.ops[0], ast.NotIn):
-                return self.visit(ast.UnaryOp(op=ast.Not(), operand=dunder_node))
             return dunder_node
         typed_cmp = copy(node)
         typed_cmp.left = self.visit(node.left)
@@ -887,6 +1017,16 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
             # We need the function type inside for recursion
             self.set_variable_type(node.name, tfd.typ)
             tfd.body = self.visit_sequence(node.body)
+            # Its possible that bound_variables might have changed after visiting body
+            bv = {
+                v: self.variable_type(v)
+                for v in externally_bound_vars(node)
+                if not v in ["List", "Dict"]
+            }
+            if bv != tfd.typ.typ.bound_vars:
+                # node was modified in place, so we can simply rerun visit_FunctionDef
+                self.exit_scope()
+                return self.visit_FunctionDef(node)
             # Check that return type and annotated return type match
             rets_extractor = ReturnExtractor(functyp.rettyp)
             rets_extractor.check_fulfills(tfd)
@@ -927,9 +1067,10 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
             values = []
             prevtyps = {}
             for e in node.values:
-                values.append(self.visit(e))
+                e_visited = self.visit(e)
+                values.append(e_visited)
                 typchecks, _ = TypeCheckVisitor(self.allow_isinstance_anything).visit(
-                    values[-1]
+                    e_visited
                 )
                 # for the time after the shortcut and the variable type to the specialized type
                 wrapped = self.implement_typechecks(typchecks)
@@ -960,7 +1101,7 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
             self.implement_typechecks(prevtyps)
             tt.values = values
         else:
-            tt.values = [self.visit(e) for e in node.values]
+            raise NotImplementedError(f"Boolean operator {node.op} not supported")
         tt.typ = BoolInstanceType
         assert all(
             BoolInstanceType >= e.typ for e in tt.values
@@ -970,10 +1111,7 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
     def visit_UnaryOp(self, node: UnaryOp) -> TypedUnaryOp:
         dunder_node = self.dunder_override(node)
         if dunder_node is not None:
-            if isinstance(node.op, ast.Not):
-                node.operand = dunder_node
-            else:
-                return dunder_node
+            return dunder_node
         tu = copy(node)
         tu.operand = self.visit(node.operand)
         tu.typ = tu.operand.typ.typ.unop_type(node.op).rettyp
@@ -996,27 +1134,32 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
             assert (
                 ts.value.typ.typ.typs
             ), "Accessing elements from the empty tuple is not allowed"
+            if isinstance(ts.slice, UnaryOp) and isinstance(ts.slice.op, USub):
+                ts.slice = self.visit(Constant(-ts.slice.operand.value))
             if isinstance(ts.slice, Constant) and isinstance(ts.slice.value, int):
                 assert ts.slice.value < len(
                     ts.value.typ.typ.typs
                 ), f"Subscript index out of bounds for tuple. Accessing index {ts.slice.value} in tuple with {len(ts.value.typ.typ.typs)} elements ({ts.value.typ.python_type()})"
                 ts.typ = ts.value.typ.typ.typs[ts.slice.value]
-            elif all(ts.value.typ.typ.typs[0] == t for t in ts.value.typ.typ.typs):
-                ts.typ = ts.value.typ.typ.typs[0]
             else:
                 raise TypeInferenceError(
-                    f"Could not infer type of subscript of typ {ts.value.typ.typ.__class__}"
+                    f"Could not infer type of subscript of typ {ts.value.typ.python_type()}"
                 )
         elif isinstance(ts.value.typ.typ, PairType):
+            if isinstance(ts.slice, UnaryOp) and isinstance(ts.slice.op, USub):
+                ts.slice = self.visit(Constant(-ts.slice.operand.value))
             if isinstance(ts.slice, Constant) and isinstance(ts.slice.value, int):
+                assert (
+                    -3 < ts.slice.value < 2
+                ), f"Can only access -2, -1, 0 or 1 index in pairs, found {ts.slice.value}"
                 ts.typ = (
                     ts.value.typ.typ.l_typ
-                    if ts.slice.value == 0
+                    if ts.slice.value % 2 == 0
                     else ts.value.typ.typ.r_typ
                 )
             else:
                 raise TypeInferenceError(
-                    f"Could not infer type of subscript of typ {ts.value.typ.typ.__class__}"
+                    f"Could not infer type of subscript of typ {ts.value.typ.python_type()}"
                 )
         elif isinstance(ts.value.typ.typ, ListType):
             if not isinstance(ts.slice, Slice):
@@ -1143,26 +1286,50 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
                     "If you are certain of what you are doing, please use the flag '--allow-isinstance-anything'."
                 )
             tc.typechecks = TypeCheckVisitor(self.allow_isinstance_anything).visit(tc)
-        try:
-            tc.func = self.visit(node.func)
-        except Exception as e:
-            # might be a method, duck test for class_name, method_name should
-            try:
-                func_variable_type = self.variable_type(tc.func.value.id)
-                class_name = func_variable_type.typ.record.name
-                method_name = f"{class_name}_{tc.func.attr}"
+
+        # Check for expanded Union funcs
+        if isinstance(node.func, ast.Name):
+            expanded_unions = {
+                k: v
+                for scope in self.scopes
+                for k, v in scope.items()
+                if k.startswith(f"{node.func.orig_id}+")
+            }
+            for k, v in expanded_unions.items():
+                argtyps = v.typ.argtyps
+                if len(tc.args) != len(argtyps):
+                    continue
+                for a, ap in zip(tc.args, argtyps):
+                    if ap != a.typ:
+                        break
+                else:
+                    node.func = ast.Name(
+                        id=k, orig_id=f"unknown orig_id for {k}", ctx=ast.Load()
+                    )
+                    break
+
+        subbed_method = False
+        if isinstance(tc.func, Attribute):
+            # might be a method, test whether the variable is a record and if the method exists
+            accessed_var = self.visit(tc.func.value)
+            if (
+                isinstance(accessed_var.typ, InstanceType)
+                and isinstance(accessed_var.typ.typ, RecordType)
+                and tc.func.attr != "to_cbor"
+            ):
+                class_name = accessed_var.typ.typ.record.name
+                method_name = f"{class_name}_+_{tc.func.attr}"
                 # If method_name found then use this.
-                self.variable_type(method_name)
-            except Exception:
-                # if this fails raise original error
-                raise e
-            n = ast.Name(id=method_name, ctx=ast.Load())
-            n.orig_id = node.func.attr
-            tc.func = self.visit(n)
-            tc.func.orig_id = node.func.attr
-            c_self = ast.Name(id=node.func.value.id, ctx=ast.Load())
-            c_self.orig_id = None
-            tc.args.insert(0, self.visit(c_self))
+                if self.is_defined_in_current_scope(method_name):
+                    n = ast.Name(id=method_name, ctx=ast.Load())
+                    n.orig_id = node.func.attr
+                    tc.func = self.visit(n)
+                    tc.func.orig_id = node.func.attr
+                    tc.args.insert(0, accessed_var)
+                    subbed_method = True
+
+        if not subbed_method:
+            tc.func = self.visit(node.func)
 
         # might be a class
         if isinstance(tc.func.typ, ClassType):
@@ -1183,12 +1350,12 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
             functyp = tc.func.typ.typ
             assert len(tc.args) == len(
                 functyp.argtyps
-            ), f"Signature of function does not match number of arguments. Expected {len(functyp.argtyps)} arguments with these types: {functyp.argtyps} but got {len(tc.args)} arguments."
+            ), f"Signature of function does not match number of arguments. Expected {len(functyp.argtyps)} arguments but got {len(tc.args)} arguments."
             # all arguments need to be subtypes of the parameter type
             for i, (a, ap) in enumerate(zip(tc.args, functyp.argtyps)):
                 assert (
                     ap >= a.typ
-                ), f"Signature of function does not match arguments in argument {i}. Expected this type: {ap} but got {a.typ}."
+                ), f"Signature of function does not match arguments in argument {i}. Expected this type: {ap.python_type()} but got {a.typ.python_type()}."
             tc.typ = functyp.rettyp
             return tc
         raise TypeInferenceError("Could not infer type of call")
@@ -1268,24 +1435,18 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
         new_g = copy(g)
         if isinstance(g.target, Tuple):
             raise NotImplementedError(
-                "Type deconstruction in for loops is not supported yet"
+                "Type deconstruction in comprehensions is not supported yet"
             )
         new_g.iter = self.visit(g.iter)
         itertyp = new_g.iter.typ
         assert isinstance(
             itertyp, InstanceType
         ), "Can only iterate over instances, not classes"
-        if isinstance(itertyp.typ, TupleType):
-            assert itertyp.typ.typs, "Iterating over an empty tuple is not allowed"
-            vartyp = itertyp.typ.typs[0]
-            assert all(
-                itertyp.typ.typs[0] == t for t in new_g.iter.typ.typs
-            ), "Iterating through a tuple requires the same type for each element"
-        elif isinstance(itertyp.typ, ListType):
+        if isinstance(itertyp.typ, ListType):
             vartyp = itertyp.typ.typ
         else:
             raise NotImplementedError(
-                "Type inference for loops over non-list objects is not supported"
+                "Iterating over non-list objects is not (yet) supported"
             )
         self.set_variable_type(g.target.id, vartyp)
         new_g.target = self.visit(g.target)
@@ -1298,8 +1459,26 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
         self.enter_scope()
         # first evaluate generators for assigned variables
         typed_listcomp.generators = [self.visit(s) for s in node.generators]
-        # then evaluate elements
+
+        # collect isinstance type narrowing from all conditions in all generators
+        all_typechecks = {}
+        for gen in typed_listcomp.generators:
+            for if_expr in gen.ifs:
+                typchecks, _ = TypeCheckVisitor(self.allow_isinstance_anything).visit(
+                    if_expr
+                )
+                all_typechecks.update(typchecks)
+
+        # apply type narrowing before evaluating the element
+        wrapped = self.implement_typechecks(all_typechecks)
+        self.wrapped.extend(wrapped.keys())
+
+        # then evaluate elements with narrowed types
         typed_listcomp.elt = self.visit(node.elt)
+
+        # clean up wrapped variables
+        self.wrapped = [x for x in self.wrapped if x not in wrapped.keys()]
+
         self.exit_scope()
         typed_listcomp.typ = InstanceType(ListType(typed_listcomp.elt.typ))
         return typed_listcomp
@@ -1310,9 +1489,27 @@ class AggressiveTypeInferencer(CompilingNodeTransformer):
         self.enter_scope()
         # first evaluate generators for assigned variables
         typed_dictcomp.generators = [self.visit(s) for s in node.generators]
-        # then evaluate elements
+
+        # collect isinstance type narrowing from all conditions in all generators
+        all_typechecks = {}
+        for gen in typed_dictcomp.generators:
+            for if_expr in gen.ifs:
+                typchecks, _ = TypeCheckVisitor(self.allow_isinstance_anything).visit(
+                    if_expr
+                )
+                all_typechecks.update(typchecks)
+
+        # apply type narrowing before evaluating the elements
+        wrapped = self.implement_typechecks(all_typechecks)
+        self.wrapped.extend(wrapped.keys())
+
+        # then evaluate elements with narrowed types
         typed_dictcomp.key = self.visit(node.key)
         typed_dictcomp.value = self.visit(node.value)
+
+        # clean up wrapped variables
+        self.wrapped = [x for x in self.wrapped if x not in wrapped.keys()]
+
         self.exit_scope()
         typed_dictcomp.typ = InstanceType(
             DictType(typed_dictcomp.key.typ, typed_dictcomp.value.typ)
@@ -1360,17 +1557,17 @@ class RecordReader(NodeVisitor):
         self.attributes = []
         self._type_inferencer = type_inferencer
 
-    @classmethod
-    def extract(cls, c: ClassDef, type_inferencer: AggressiveTypeInferencer) -> Record:
-        f = cls(type_inferencer)
-        f.visit(c)
-        if f.constructor is None:
+    def extract(self, c: ClassDef) -> Record:
+        self.visit(c)
+        if self.constructor is None:
             det_string = RecordType(
-                Record(f.name, f.orig_name, 0, frozenlist(f.attributes))
+                Record(self.name, self.orig_name, 0, frozenlist(self.attributes))
             ).pluthon_type(skip_constructor=True)
             det_hash = sha256(str(det_string).encode("utf8")).hexdigest()
-            f.constructor = int(det_hash, 16) % 2**32
-        return Record(f.name, f.orig_name, f.constructor, frozenlist(f.attributes))
+            self.constructor = int(det_hash, 16) % 2**32
+        return Record(
+            self.name, self.orig_name, self.constructor, frozenlist(self.attributes)
+        )
 
     def visit_AnnAssign(self, node: AnnAssign) -> None:
         assert isinstance(
@@ -1391,7 +1588,7 @@ class RecordReader(NodeVisitor):
                 )
             )
             return
-        assert typ == IntegerType, "CONSTR_ID must be assigned an integer"
+        assert typ == IntegerType(), "CONSTR_ID must be assigned an integer"
         assert isinstance(
             node.value, Constant
         ), "CONSTR_ID must be assigned a constant integer"
@@ -1434,10 +1631,6 @@ class RecordReader(NodeVisitor):
         raise NotImplementedError(f"Can not compile {ast.dump(node)} inside of a class")
 
 
-def typed_ast(ast: AST):
-    return AggressiveTypeInferencer().visit(ast)
-
-
 def map_to_orig_name(name: str):
     return re.sub(r"_\d+$", "", name)
 
